/*
  Gypsy Road - GUI Layout
  Copyright (C) 2022-2025  Karl Robillard
  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include "event.h"
#include "gypsy_types.h"
#include "stringTable.h"
#include "txf_draw.h"

enum GuiOpcode {
    GUI_END,

    // Position Pen
    PEN_PER,        // percent-x, percent-y

    // Layout
    LAYOUT_H,
    LAYOUT_V,
    LAYOUT_GRID,    // columns
    LAYOUT_END,
    MARGIN_PER,     // percent
    MARGIN_H_PER,   // percent
    MARGIN_V_PER,   // percent
    SPACING_PER,    // percent
    SPACING_EM,     // font-em-tenth
    FIX_WIDTH_PER,  // percent
    FIX_HEIGHT_PER, // percent
    FIX_WIDTH_EM,   // font-em-tenth
    FIX_HEIGHT_EM,  // font-em-tenth
    FROM_BOTTOM,
    FROM_RIGHT,
    ALIGN_LEFT,
    ALIGN_RIGHT,
    ALIGN_TOP,
    ALIGN_BOTTOM,
    ALIGN_H_CENTER,
    ALIGN_V_CENTER,
    ALIGN_CENTER,
    GAP_PER,        // percent
    GRID_EMPTY,

    // Drawing
    FONT_N,         // font-index
    FONT_SIZE,      // point-size
    FONT_VSIZE,     // scaled-vga-height (480 pixels)
    FONT_COLOR,     // color-index
    BG_COLOR_CI,    // color-index

    // Widgets
    ARRAY_DT_AREA,  // initial-wid, DATA GuiArea*
    CHECK_DT_U32,   // DATA uint32_t*
    BUTTON_DT_S,    // DATA const char*     STORE_AREA
    CHECKBOX,       // bit-number           STORE_AREA
    LABEL_DT_S,     // DATA const char*
    LIST_DIM,       // columns, rows        STORE_AREA
    SLIDER_DT_F,    // tdi, DATA float*     STORE_AREA
    MIN_WIDTH_EM,   // font-em
    STORE_DT_AREA,  // DATA int16_t[4]
    STORE_AREA
};

enum GuiWidgetType {
    GUI_WT_NONE,
    GUI_WT_BUTTON,
    GUI_WT_CHECKBOX,
    GUI_WT_LABEL,
    GUI_WT_LIST,
    GUI_WT_LIST_ITEM,
    GUI_WT_MENU,
    GUI_WT_MENU_ITEM,
    GUI_WT_SLIDER
};

typedef struct {
    int16_t x, y, w, h;
} GuiRect;

// Matches BTree2Box
typedef struct {
    int16_t x, y;
    int16_t x2, y2;
    uint8_t wtype;      // GuiWidgetType
    uint8_t tdi;        // Type-specific data identity
    int16_t wid;        // Area index
} GuiArea;

typedef struct {
    float   tabStop;
    float   fontScale;
    uint8_t color;
    uint8_t selColor;
} ListCellStyle;

typedef struct {
    TxfDrawState ds;
    float   psizeList;
    uint8_t selected;
    uint8_t tabCount;
    ListCellStyle cell[4];
} ListDrawState;

typedef struct {
    float pos, target;
} ScrollPos;

typedef struct {
    struct GuiProgram* prog;
    GuiArea* gbox;
    uint16_t gboxUsed;
    uint16_t gboxAvail;
    float    actValue;      // Slider value.
    uint16_t flags;
    uint16_t buttonMode;
    int16_t  buttonDown;    // gbox index or WID_NONE.
    int16_t  selItem;
    int16_t  selItemRend;
    uint16_t popupX;
    uint16_t popupY;
    int16_t  tipId;
    uint32_t tipTimer;
    uint16_t tipX;
    uint16_t tipY;
} GuiState;


void   gui_setStyle(const char* texName);
float* gui_layout(float* attr, const GuiRect* root, TxfDrawState*,
                  const uint8_t* bytecode, GuiArea**, const void** data);
float* gui_emitListItems(float* attr, ListDrawState*, StringTable* st,
                         int select);
void   gui_animateScroll(ScrollPos*);
float* gui_emitQuadCi(float* attr, const float* rect, float colorIndex);
float* gui_emitText(TxfDrawState*, float* attr, const char* text, uint32_t len);
void*  gui_areaTree(const GuiArea* areas, int count);
const GuiArea* gui_pick(const void* tree, const GuiArea* areas,
                        uint16_t x, uint16_t y);

struct WorkBuffer;
struct Stage;

void guis_init(GuiState*, Symbol progName, GuiArea* areas, uint16_t count);
void guis_layout(GuiState*, struct WorkBuffer* work, int rol);
const GuiArea* guis_mouseButton(GuiState*, const InputEvent*);
const GuiArea* guis_mouseMove(GuiState*, const InputEvent*);
void guis_popupEnter(struct Stage*, uintptr_t args);
int  guis_popupDispatch(struct Stage*, const InputEvent* ev);
float* guis_emitPopupMenu(GuiState*, float* attr, TxfDrawState* ds,
                          const char** items, int itemCount, uint32_t mask);
