/*
 * gypsy.h
 */

#include <stdbool.h>
#include "anim.h"
#include "notify.h"
#include "present.h"
#include "screen.h"
#include "stage.h"
#include "stringTable.h"

enum NotifySender {
    // Sender Id           Message
    SENDER_GAME,        // int* (GameMessageType)
    SENDER_SETTINGS,    // int* (settings location)
    SENDER_DISPLAY,     // NULL or ScreenState*
    SENDER_NEW_LOCALE,  // Locale*
#ifdef USE_XU4
    SENDER_LOCATION,    // MoveEvent*
    SENDER_PARTY,       // PartyEvent*
    SENDER_AURA,        // Aura*
    SENDER_MENU         // MenuEvent*
#endif
};

enum GameMessageType {
    GAME_INIT,
    GAME_FREE,
    GAME_RESET,
    GAME_SAVE,
    GAME_LOAD
};

#define GAME_MESSAGE_TYPE   *((int*) message)

// All games must have these stages.
enum StandardStageId {
    STD_STAGE_INTRO,
    STD_STAGE_PLAY,
    STD_STAGE_BROWSE
};

typedef struct FrameSleep FrameSleep;
typedef struct MainLoop MainLoop;
typedef struct GameServices GameServices;

#define FS_SAMPLES  8

struct FrameSleep {
    ClockMs  frameInterval;     // Milliseconds between display updates.
    ClockMs  realTime;
    int ftime[FS_SAMPLES];      // Msec elapsed for past frames.
    int ftimeSum;
    uint16_t ftimeIndex;
    uint16_t fsleep;            // Msec to sleep - adjusted slowly.
};

struct MainLoop {
    FrameSleep fs;
    ClockMs  runTime;
    uint16_t stageEvents;
};

struct ImageInfo;
struct TileMap;

struct GameServices {
    StringTable resourcePaths;
    uint8_t ufileLocation[8];   // Cache for u4file_paths.
    NotifyBus notifyBus;
    MainLoop loop;
    SettingsBase* settings;
    Config* config;
    ScreenBase* screen;
    void* gpu;
    PresentProgram present;
    Animator varyRt;            // Real time
    Animator varySt;            // Simulation time
    struct ImageInfo* textureInfo;  // Stores ImageRect of loaded textures.
    struct TileMap* tilemapBank;
    void* gameSim;              // Game world simulation.
    //const char* errorMessage;
    uint16_t textureBankSize;
    uint16_t tilemapBankSize;
    uint16_t gameReset;         // Load another game.
    ClockMs  timerInterval;     // Milliseconds between timedEvents ticks.
    ClockMs  rclock;            // Global "realtime" clock (milliseconds).
    uint32_t randomFx[17];      // Effects random number generator state.
    uint8_t exitCode;
    bool verbose;
};

extern GameServices gs;

#define gs_listen(msk,func,user)    notify_listen(&gs.notifyBus,msk,func,user)
#define gs_unplug(id)               notify_unplug(&gs.notifyBus,id)
#define gs_emitMessage(sid,data)    notify_emit(&gs.notifyBus,sid,data);

bool gs_fileExists(const char* fname);
int  gs_filePath(const char* fname, const char* ext,
                 char* path, size_t pathSize);
void gs_openUrl(const char* url);
void gs_recordKey(const InputEvent*);
void gs_srandom(uint32_t seed);
int  gs_random(int upperval);
int  gs_randomFx(int upperval);
uint32_t gs_randomFx32();

void frameSleepInit(FrameSleep* fs, int frameDuration);
void mainLoop(MainLoop*);

#define image_texInfo(slot)     (gs.textureInfo + (slot))
void image_initTextureInfo();
void image_freeTextureInfo();

void tilemap_initBank();
void tilemap_freeBank();
