/*
  Gypsy Road - Stage State Machine
  Copyright (C) 2022-2024  Karl Robillard
  SPDX-License-Identifier: GPL-3.0-or-later
*/

#ifndef STAGE_H
#define STAGE_H

#include "event.h"

#define STAGE_FLAG_CHILD    1
#define STAGE_FLAG_ROOT     2
#define STAGE_FLAG_FRAME    4
#define STAGE_FLAG_USER     0xff00

#define STAGE_EX_DONE   0x7ffd
#define STAGE_EX_ABORT  0x7ffe
#define STAGE_EX_EXIT   0x7fff

#define STAGE_SEND_CYCLE    1
#define STAGE_SEND_RESIZE   2
#define STAGE_TOGGLE_FS     4

#define stage_done()    stage_runS(st, STAGE_EX_DONE)
#define stage_abort()   stage_runS(st, STAGE_EX_ABORT)
#define stage_exit()    stage_runG(STAGE_EX_EXIT, 0)
#define stage_run(id)   stage_runS(st, id)
#define stage_runArgs(id,arg)   stage_runSArgs(st, id, arg)

typedef struct StageDef StageDef;
typedef struct Stage Stage;

struct StageDef {
    void    (*enter) (Stage*, uintptr_t);
    void    (*leave) (Stage*);
    int     (*dispatch) (Stage*, const InputEvent*);
    uint16_t flags;
    uint16_t tickInterval;
};

struct Stage {
    const StageDef* def;
    int (*dispatch) (Stage*, const InputEvent*);
    void* data;
    uint32_t msecInterval;
    uint32_t msecTimeout;
    uint16_t tickInterval;
    uint16_t tickCount;
    uint16_t dstate;        // User defined (dispatch state).
    uint16_t _pad;
};

void         stage_init(const StageDef* stages);
int          stage_id(const Stage*);
int          stage_currentId();
void         stage_runG(int id, uintptr_t args);
void         stage_runS(Stage*, int id);
void         stage_runSArgs(Stage*, int id, uintptr_t args);
void         stage_pushDispatch(int id, void* data);
void         stage_popDispatch(Stage*);
const Stage* stage_transition();
Stage*       stage_current();
int          stage_exception();
void         stage_startMsecTimer(Stage*, uint32_t millisec);
void         stage_stopMsecTimer(Stage*);
void         stage_tick(int send);

void  nop_enter(Stage*, uintptr_t);
void  nop_leave(Stage*);
int   nop_dispatch(Stage*, const InputEvent*);
void  iconify_enter(Stage*, uintptr_t);

#endif  // STAGE_H
