/*
 * Gypsy Example - Main
 * Copyright (C) 2023-2025  Karl Robillard
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "error.h"
#include "gui.h"
#include "gypsy.h"
#include "sound.h"
#include "stage.h"

#define GAME_MODULE "data.mod"
#define SOUNDTRACK  NULL
#define SCREEN_RW   640
#define SCREEN_RH   480
#define UPDATE_MSEC (1000 / 30)

#define APP_NAME    "Gypsy Example"
#define APP_PATH    "gypsy-example"
#define APP_VERSION "0.1"

extern void game_notice(int, void*, void*);
extern const StageDef stages[];

#ifdef ANDROID
const char* androidInternalData();
#endif

#ifdef _WIN32
void redirectIOToConsole();
#endif

#ifdef DEBUG
#include "irecord.h"
extern InputRecorder gInputRec;
#endif

#include "app_settings.c"


enum OptionsFlag {
    OPT_FULLSCREEN = 1,
    OPT_NO_INTRO   = 2,
    OPT_NO_AUDIO   = 4,
    OPT_VERBOSE    = 8,
    OPT_RECORD     = 0x10,
    OPT_REPLAY     = 0x20,
    OPT_STAGE      = 0x40,
    OPT_TEST_SAVE  = 0x80
};

typedef struct {
    uint16_t flags;
    uint16_t used;
    uint16_t scale;
    uint16_t stage;
    const char* filter;
    const char* module;
    const char* profile;
    const char* recordFile;
} Options;

#define strEqual(A,B)       (strcmp(A,B) == 0)
#define strEqualAlt(A,B,C)  (strEqual(A,B) || strEqual(A,C))

/*
 * Return exit status code or -1 if the program should continue.
 */
static int parseOptions(Options* opt, int argc, char** argv)
{
    int i;
    for (i = 0; i < argc; i++) {
        if (strEqual(argv[i], "--filter"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->filter = argv[i];
        }
        else if (strEqualAlt(argv[i], "-s", "--scale"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->scale = atoi(argv[i]);
        }
#ifdef CONF_MODULE
        else if (strEqualAlt(argv[i], "-m", "--module"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->module = argv[i];
        }
#endif
        else if (strEqualAlt(argv[i], "-p", "--profile"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->profile = argv[i];
        }
        else if (strEqualAlt(argv[i], "-i", "--skip-intro"))
        {
            opt->flags |= OPT_NO_INTRO;
            opt->used  |= OPT_NO_INTRO;
        }
        else if (strEqualAlt(argv[i], "-v", "--verbose"))
        {
            opt->flags |= OPT_VERBOSE;
            opt->used  |= OPT_VERBOSE;
        }
        else if (strEqualAlt(argv[i], "-f", "--fullscreen"))
        {
            opt->flags |= OPT_FULLSCREEN;
            opt->used  |= OPT_FULLSCREEN;
        }
        else if (strEqualAlt(argv[i], "-q", "--quiet"))
        {
            opt->flags |= OPT_NO_AUDIO;
            opt->used  |= OPT_NO_AUDIO;
        }
        else if (strEqualAlt(argv[i], "-h", "--help"))
        {
            printf(APP_NAME "\nv" APP_VERSION " (%s)\n\n", __DATE__);
            printf(
            "Options:\n"
            "      --filter <string>   Specify display filtering mode.\n"
            "                          (point, HQX, xBR-lv2, xBRZ)\n"
            "  -f, --fullscreen        Run in fullscreen mode.\n"
            "  -h, --help              Print this message and quit.\n"
            "  -i, --skip-intro        Skip the intro. and load the last saved game.\n"
#ifdef CONF_MODULE
            "  -m, --module <file>     Specify game module (default is Ultima-IV).\n"
#endif
            "  -p, --profile <string>  Use another set of settings and save files.\n"
            "  -q, --quiet             Disable audio.\n"
            "  -s, --scale <int>       Specify display scaling factor (1-5).\n"
            "  -v, --verbose           Enable verbose console output.\n"
#ifdef DEBUG
            "\nDEBUG Options:\n"
            "  -c, --capture <file>    Record user input.\n"
            "  -r, --replay <file>     Play using recorded input.\n"
            "      --stage <int>       Start game in given stage (default 0).\n"
            "      --test-save         Save to /tmp/xu4/ and quit.\n"
#endif
            "\nHomepage: http://xu4.sourceforge.net\n");

            return EX_OK;
        }
#ifdef DEBUG
        else if (strEqualAlt(argv[i], "-c", "--capture"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->recordFile = argv[i];
            opt->flags |= OPT_RECORD;
            opt->used  |= OPT_RECORD;
        }
        else if (strEqualAlt(argv[i], "-r", "--replay"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->recordFile = argv[i];
            opt->flags |= OPT_REPLAY;
            opt->used  |= OPT_REPLAY;
        }
        else if (strEqual(argv[i], "--stage"))
        {
            if (++i >= argc)
                goto missing_value;
            opt->stage = atoi(argv[i]);
            opt->flags |= OPT_STAGE;
        }
        else if (strEqual(argv[i], "--test-save"))
        {
            opt->flags |= OPT_TEST_SAVE;
        }
#endif
        else {
            errorFatal(EX_USAGE, "Unrecognized argument: %s\n\n"
                "Use --help for a list of supported arguments.", argv[i]);
            return EX_USAGE;
        }
    }
    return -1;

missing_value:
    errorFatal(EX_USAGE, "%s requires a value. See --help for more detail.",
               argv[i-1]);
    return EX_USAGE;
}

//----------------------------------------------------------------------------

static void initResourcePaths(StringTable* st)
{
#ifdef ANDROID
    sst_init(st, 2, 64);
    sst_append(st, androidInternalData(), -1);
#else
    char* home;

    sst_init(st, 4, 32);

    sst_append(st, ".", 1);
#ifdef _WIN32
    home = getenv("LOCALAPPDATA");
    if (home && home[0])
        sst_appendCon(st, home, "\\" APP_PATH);
#else
    home = getenv("HOME");
    if (home && home[0]) {
#ifdef __APPLE__
        sst_appendCon(st, home, "/Library/Application Support/" APP_PATH);
#else
        sst_appendCon(st, home, "/.local/share/" APP_PATH);
#endif
    }
    sst_append(st, "/usr/share/" APP_PATH, -1);
    sst_append(st, "/usr/local/share/" APP_PATH, -1);
#endif
#endif
}

#ifdef DEBUG
static void servicesFree(GameServices*);
#endif

int servicesInit(GameServices* gs, Options* opt) {
    uint16_t sloc[5];

    gs->verbose = opt->flags & OPT_VERBOSE;

    initResourcePaths(&gs->resourcePaths);

    /* Setup the message bus early to make it available to other services. */
    notify_init(&gs->notifyBus, 8);

    /* Initialize the settings and update with any command-line arguments. */
    gs->settings = settings_init(APP_NAME, APP_PATH, APP_VERSION,
                                 settings_data, sizeof(settings_data));
    if (! gs->settings)
        return gs->exitCode;

    settings_location(sloc, 5, "fullscreen scale screen-filter\n"
                               "game soundtrack");

    if (opt->used & OPT_FULLSCREEN)
        settings_pokeB(sloc[0], opt->flags & OPT_FULLSCREEN);
    if (opt->scale)
        settings_pokeI(sloc[1], opt->scale);
    if (opt->filter)
        settings_pokeS(sloc[2], opt->filter);
    gs->settings->dirty = 0;    // Options are temporary.

    gs->config = conf_init(GAME_MODULE, SOUNDTRACK);
    if (! gs->config)
        return gs->exitCode;

    image_initTextureInfo();
#ifdef USE_TILEMAP
    tilemap_initBank();
#endif

    gs->screen = screen_init(SCR_FORMAT_SCALED, SCREEN_RW, SCREEN_RH,
                             gs->settings);
    if (! gs->screen)
        return gs->exitCode;

    if (! (opt->flags & OPT_NO_AUDIO))
        sound_init(conf_bankSize(RT_SOUND));

    anim_init(&gs->varyRt, 32, NULL, NULL);
    anim_init(&gs->varySt, 32, NULL, NULL);
    gui_setStyle("tex-gui");

    {
    uint32_t seed;

#ifdef DEBUG
    irec_init(&gInputRec);

    if (opt->flags & OPT_REPLAY) {
        seed = irec_replay(&gInputRec, opt->recordFile);
        if (! seed) {
            servicesFree(gs);
            errorFatal(EX_NOINPUT, "Cannot open recorded input from %s",
                       opt->recordFile);
            return gs->exitCode;
        }
    } else if (opt->flags & OPT_RECORD) {
        seed = time(NULL);
        if (! irec_beginRecording(&gInputRec, opt->recordFile, seed)) {
            servicesFree(gs);
            errorFatal(EX_CANTCREAT, "Cannot open recording file %s",
                       opt->recordFile);
            return gs->exitCode;
        }
    } else
#endif
        seed = time(NULL);

    gs_srandom(seed);
    }

    notify_listen(&gs->notifyBus, 1<<SENDER_GAME | 1<<SENDER_NEW_LOCALE,
                  game_notice, gs);
    stage_init(stages);

#ifdef DEBUG
    if (opt->flags & OPT_STAGE)
        stage_runG(opt->stage, 0);
    else
#endif
    stage_runG((opt->flags & OPT_NO_INTRO) ? STD_STAGE_PLAY
                                           : STD_STAGE_INTRO, 0);
    return gs->exitCode;
}

static void servicesFreeGame(GameServices* gs) {
    present_free(&gs->present);
    anim_free(&gs->varyRt);
    anim_free(&gs->varySt);
#ifdef USE_TILEMAP
    tilemap_freeBank();
#endif
    image_freeTextureInfo();
    sound_free();
    screen_free(gs->screen);
    conf_free(gs->config);
}

static void servicesFree(GameServices* gs) {
#ifdef DEBUG
    irec_endRecording(&gInputRec);
#endif

    servicesFreeGame(gs);

    settings_free(gs->settings);
    notify_free(&gs->notifyBus);
    //u4fcleanup();
    sst_free(&gs->resourcePaths);
}

static void servicesReset(GameServices* gs) {
    servicesFreeGame(gs);

    uint32_t seed = time(NULL);
    gs_srandom(seed);

    stage_runG(0, 0);
}

static void gameNotify(int gameMsg)
{
    gs_emitMessage(SENDER_GAME, &gameMsg);
}

int main(int argc, char *argv[])
{
#ifdef _WIN32
    redirectIOToConsole();
#endif

    {
    Options opt;
    int status;

    /* Parse arguments before setup in case the user only wants some help. */
    memset(&opt, 0, sizeof opt);
    status = parseOptions(&opt, argc-1, argv+1);
    if (status >= 0)
        return status;

    memset(&gs, 0, sizeof(gs));
    status = servicesInit(&gs, &opt);
    if (status != EX_OK)
        return status;
    }

    gameNotify(GAME_INIT);

    while (conf_loadInitialResources())
    {
        gs.loop.runTime = 0;
        gs.timerInterval = UPDATE_MSEC;
        frameSleepInit(&gs.loop.fs, gs.timerInterval);

        mainLoop(&gs.loop);

        if (! gs.gameReset)
            break;

        gameNotify(GAME_RESET);
        gs.gameReset = 0;
        gs.rclock = 0;
        servicesReset(&gs);
    }

    gameNotify(GAME_FREE);
    servicesFree(&gs);
    return gs.exitCode;
}
